from . import conda
from .conda.conda_flow_decorator import CondaFlowDecorator
from .conda.conda_step_decorator import CondaStepDecorator
from .conda.conda_environment import CondaEnvironment

FLOW_DECORATORS = [CondaFlowDecorator]
STEP_DECORATORS = [CondaStepDecorator]
ENVIRONMENTS = [CondaEnvironment]

def get_plugin_cli():
    from . import environment_cli
    return [environment_cli.cli]

__mf_promote_submodules__ = ["conda"]



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false
import os
import shutil
import sys
import time
from typing import Any, Callable

from metaflow.metaflow_config import DATASTORE_LOCAL_DIR, CONDA_MAGIC_FILE_V2

from metaflow.cli import echo_always

from metaflow.plugins.env_escape import generate_trampolines, ENV_ESCAPE_PY

from .env_descr import read_conda_manifest
from .conda import Conda


def my_echo_always(*args: Any, **kwargs: Any) -> Callable[..., None]:
    kwargs["err"] = False
    return echo_always(*args, **kwargs)


def bootstrap_environment(
    flow_name: str, step_name: str, req_id: str, full_id: str, datastore_type: str
):
    start = time.time()
    my_echo_always("    Setting up Conda ...", nl=False)
    my_conda = Conda(my_echo_always, datastore_type, mode="remote")
    setup_conda_manifest()
    my_echo_always(" done in %d seconds." % int(time.time() - start))

    cached_info = read_conda_manifest(DATASTORE_LOCAL_DIR)
    resolved_env = cached_info.env_for(req_id, full_id)
    if resolved_env is None:
        raise RuntimeError(
            "Cannot find cached environment for hash %s:%s" % (req_id, full_id)
        )
    # Install the environment; this will fetch packages as well.
    my_conda.create_for_step(step_name, resolved_env, do_symlink=True)

    # Setup anything needed by the escape hatch
    if ENV_ESCAPE_PY is not None:
        cwd = os.getcwd()
        generate_trampolines(cwd)
        # print("Environment escape will use %s as the interpreter" % ENV_ESCAPE_PY)
    else:
        pass
        # print("Could not find a environment escape interpreter")


def setup_conda_manifest():
    manifest_folder = os.path.join(os.getcwd(), DATASTORE_LOCAL_DIR)
    if not os.path.exists(manifest_folder):
        os.makedirs(manifest_folder)
    shutil.move(
        os.path.join(os.getcwd(), CONDA_MAGIC_FILE_V2),
        os.path.join(manifest_folder, CONDA_MAGIC_FILE_V2),
    )


if __name__ == "__main__":
    bootstrap_environment(*sys.argv[1:])



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false

from itertools import chain
import json
import os
import time
import tarfile

from concurrent.futures import ThreadPoolExecutor, as_completed
from io import BytesIO
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    cast,
)

from metaflow.plugins.datastores.local_storage import LocalStorage
from metaflow.debug import debug
from metaflow.flowspec import FlowSpec

from metaflow.metaflow_config import (
    CONDA_MAGIC_FILE_V2,
    CONDA_PREFERRED_FORMAT,
)

from metaflow.metaflow_environment import MetaflowEnvironment

from .utils import get_conda_manifest_path, plural_marker

from .env_descr import CachedEnvironmentInfo, EnvID, ResolvedEnvironment
from .conda import Conda
from .conda_step_decorator import CondaStepDecorator, get_conda_decorator


class CondaEnvironment(MetaflowEnvironment):
    TYPE = "conda"
    _filecache = None
    _conda = None

    def __init__(self, flow: FlowSpec):
        self._flow = flow

        self._conda = None  # type: Optional[Conda]

        # key: EnvID; value: dict containing:
        #  - "id": key
        #  - "steps": steps using this environment
        #  - "arch": architecture of the environment
        #  - "deps": array of requested dependencies
        #  - "channels": additional channels to search
        #  - "resolved": ResolvedEnvironment or None
        #  - "already_resolved": T/F
        self._requested_envs = {}  # type: Dict[EnvID, Dict[str, Any]]

        # A conda environment sits on top of whatever default environment
        # the user has so we get that environment to be able to forward
        # any calls we don't handle specifically to that one.
        from metaflow.plugins import ENVIRONMENTS
        from metaflow.metaflow_config import DEFAULT_ENVIRONMENT

        if DEFAULT_ENVIRONMENT == self.TYPE:
            # If the default environment is Conda itself then fallback on
            # the default 'default environment'
            self.base_env = MetaflowEnvironment(self._flow)
        else:
            self.base_env = [
                e
                for e in ENVIRONMENTS + [MetaflowEnvironment]
                if e.TYPE == DEFAULT_ENVIRONMENT
            ][0](self._flow)

    def init_environment(self, echo: Callable[..., None]):
        # Print a message for now
        echo("Bootstrapping Conda environment... (this could take a few minutes)")

        self._conda = cast(Conda, self._conda)
        for step in self._flow:
            # Figure out the environments that we need to resolve for all steps
            # We will resolve all unique environments in parallel
            step_conda_dec = get_conda_decorator(self._flow, step.__name__)
            my_arch_env_id = step_conda_dec.env_id
            requested_archs = step_conda_dec.requested_architectures
            # First look up the env_id (it's the default one)
            resolved_env = self._conda.environment(my_arch_env_id)
            if resolved_env is not None:
                # Check to make sure it supports all requested architectures
                if not set(requested_archs).issubset(resolved_env.co_resolved_archs):
                    resolved_env = None
            for arch in requested_archs:
                env_id = EnvID(
                    req_id=my_arch_env_id.req_id,
                    full_id=my_arch_env_id.full_id,
                    arch=arch,
                )
                if env_id not in self._requested_envs:
                    self._requested_envs[env_id] = {
                        "id": env_id,
                        "steps": [step.name],
                        "deps": step_conda_dec.step_deps,
                        "sources": step_conda_dec.source_deps,
                        "conda_format": [CONDA_PREFERRED_FORMAT],
                        "resolved": resolved_env,
                        "already_resolved": resolved_env is not None,
                    }
                else:
                    self._requested_envs[env_id]["steps"].append(step.name)

        # At this point, we check in our backend storage if we have the files we need

        need_resolution = [
            env_id
            for env_id, req in self._requested_envs.items()
            if req["resolved"] is None
        ]
        if debug.conda:
            debug.conda_exec("Resolving environments:")
            for env_id in need_resolution:
                info = self._requested_envs[env_id]
                debug.conda_exec(
                    "%s (%s): %s" % (env_id.req_id, env_id.full_id, str(info))
                )
        if len(need_resolution):
            self._resolve_environments(echo, need_resolution)

        if self._datastore_type != "local":
            # We may need to update caches
            # Note that it is possible that something we needed to resolve, we don't need
            # to cache (if we resolved to something already cached).
            update_envs = [
                request["resolved"]
                for request in self._requested_envs.values()
                if not request["already_resolved"]
                or not request["resolved"].is_cached({"conda": request["conda_format"]})
            ]
            formats = list(
                set(
                    chain(
                        *[
                            request["conda_format"]
                            for request in self._requested_envs.values()
                        ]
                    )
                )
            )

            self._conda.cache_environments(update_envs, {"conda": formats})
        else:
            update_envs = [
                request["resolved"]
                for request in self._requested_envs.values()
                if not request["already_resolved"]
            ]
        self._conda.add_environments(update_envs)

        # Update the default environment
        for env_id, request in self._requested_envs.items():
            if env_id.full_id == "_default":
                self._conda.set_default_environment(request["resolved"].env_id)

        # We are done -- write back out the environments.
        # TODO: Not great that this is manual
        self._conda.write_out_environments()

        # Delegate to whatever the base environment needs to do.
        self.base_env.init_environment(echo)

    def validate_environment(
        self, echo: Callable[..., Any], datastore_type: str
    ) -> None:
        self._local_root = cast(str, LocalStorage.get_datastore_root_from_config(echo))
        self._datastore_type = datastore_type
        self._conda = Conda(echo, datastore_type)

        return self.base_env.validate_environment(echo, datastore_type)

    def decospecs(self) -> Tuple[str, ...]:
        # Apply conda decorator and base environment's decorators to all steps
        return ("conda",) + self.base_env.decospecs()

    def _resolve_environments(
        self, echo: Callable[..., None], env_ids: Sequence[EnvID]
    ):
        start = time.time()
        if len(env_ids) == len(self._requested_envs):
            echo(
                "    Resolving %d environment%s in flow ..."
                % (len(env_ids), plural_marker(len(env_ids))),
                nl=False,
            )
        else:
            echo(
                "    Resolving %d of %d environment%s in flow (others are cached) ..."
                % (
                    len(env_ids),
                    len(self._requested_envs),
                    plural_marker(len(self._requested_envs)),
                ),
                nl=False,
            )

        def _resolve(env_desc: Mapping[str, Any]) -> Tuple[EnvID, ResolvedEnvironment]:
            self._conda = cast(Conda, self._conda)
            env_id = cast(EnvID, env_desc["id"])
            return (
                env_id,
                self._conda.resolve(
                    env_desc["steps"],
                    env_desc["deps"],
                    env_desc["sources"],
                    env_id.arch,
                ),
            )

        self._conda = cast(Conda, self._conda)
        # NOTE: Co-resolved environments allow you to resolve a bunch of "equivalent"
        # environments for different platforms. This is great as it can allow you to
        # run code on Linux and then instantiate an environment to look at it on Mac.
        # One issue though is that the set of packages on Linux may change while those
        # on mac may not (or vice versa) so it is possible to get in the following
        # situation:
        # - Co-resolve at time A:
        #   - Get linux full_id 123 and mac full_id 456
        # - Co-resolve later at time B:
        #   - Get linux full_id 123 and mac full_id 789
        # This is a problem because now the 1:1 correspondence between co-resolved
        # environments (important for figuring out which environment to use) is broken
        #
        # To solve this problem, we consider that co-resolved environments participate
        # in the computation of the full_id (basically a concatenation of all packages
        # across all co-resolved environments). This maintains the 1:1 correspondence.
        # It has a side benefit that we can use that same full_id for all co-resolved
        # environment making one easier to find from the other (instead of using indirect
        # links)
        co_resolved_envs = (
            {}
        )  # type: Dict[str, List[Tuple[EnvID, ResolvedEnvironment]]]
        if len(env_ids):
            with ThreadPoolExecutor() as executor:
                resolution_result = [
                    executor.submit(_resolve, v)
                    for k, v in self._requested_envs.items()
                    if k in env_ids
                ]
                for f in as_completed(resolution_result):
                    env_id, resolved_env = f.result()
                    # This checks if there is the same resolved environment already
                    # cached (in which case, we don't have to check a bunch of things
                    # so makes it nicer)
                    co_resolved_envs.setdefault(env_id.req_id, []).append(
                        (env_id, resolved_env)
                    )

            # Now we know all the co-resolved environments so we can compute the full
            # ID for all those environments
            for envs in co_resolved_envs.values():
                if len(envs) > 1:
                    ResolvedEnvironment.set_coresolved_full_id([x[1] for x in envs])

                for orig_env_id, resolved_env in envs:
                    resolved_env_id = resolved_env.env_id
                    cached_resolved_env = self._conda.environment(resolved_env_id)
                    if cached_resolved_env:
                        resolved_env = cached_resolved_env
                        self._requested_envs[orig_env_id]["already_resolved"] = True

                    self._requested_envs[orig_env_id]["resolved"] = resolved_env
                    debug.conda_exec(
                        "For environment %s (%s) (deps: %s), need packages %s"
                        % (
                            orig_env_id.req_id,
                            orig_env_id.full_id,
                            ";".join(map(str, resolved_env.deps)),
                            ", ".join([p.filename for p in resolved_env.packages]),
                        )
                    )

        duration = int(time.time() - start)
        echo(" done in %d second%s." % (duration, plural_marker(duration)))

    def _get_env_id(self, step_name: str) -> Optional[EnvID]:
        conda_decorator = get_conda_decorator(self._flow, step_name)
        if conda_decorator.is_enabled():
            resolved_env = cast(Conda, self._conda).environment(conda_decorator.env_id)
            if resolved_env:
                return resolved_env.env_id
        return None

    def _get_executable(self, step_name: str) -> Optional[str]:
        env_id = self._get_env_id(step_name)
        if env_id is not None:
            # The create method in Conda() sets up this symlink when creating the
            # environment.
            return os.path.join(".", "__conda_python")
        return None

    def bootstrap_commands(self, step_name: str, datastore_type: str) -> List[str]:
        # Bootstrap conda and execution environment for step
        env_id = self._get_env_id(step_name)
        if env_id is not None:
            return [
                "export CONDA_START=$(date +%s)",
                "echo 'Bootstrapping environment ...'",
                'python -m %s.remote_bootstrap "%s" "%s" %s %s %s'
                % (
                    "metaflow_extensions.netflix_ext.plugins.conda",
                    self._flow.name,
                    step_name,
                    env_id.req_id,
                    env_id.full_id,
                    datastore_type,
                ),
                "export _METAFLOW_CONDA_ENV='%s'"
                % json.dumps(env_id).replace('"', '\\"'),
                "echo 'Environment bootstrapped.'",
                "export CONDA_END=$(date +%s)",
            ]
            # TODO: Add the PATH part (need conda directory)
        return []

    def add_to_package(self) -> List[Tuple[str, str]]:
        # TODO: Improve this to only extract the environments that we care about
        # The issue is that we need to pass a path which is a bit annoying since
        # we then can't clean it up easily.
        files = self.base_env.add_to_package()
        # Add conda manifest file to job package at the top level.
        path = get_conda_manifest_path(self._local_root)
        if os.path.exists(path):
            files.append((path, os.path.basename(path)))
        return files

    def pylint_config(self) -> List[str]:
        config = self.base_env.pylint_config()
        # Disable (import-error) in pylint
        config.append("--disable=F0401")
        return config

    def executable(self, step_name: str) -> str:
        # Get relevant python interpreter for step
        executable = self._get_executable(step_name)
        if executable is not None:
            return executable
        return self.base_env.executable(step_name)

    @classmethod
    def get_client_info(
        cls, flow_name: str, metadata: Dict[str, str]
    ) -> Dict[str, Any]:
        # TODO: FIX THIS: this doesn't work with the new format.
        if cls._filecache is None:
            from metaflow.client.filecache import FileCache

            cls._filecache = FileCache()
        info = metadata.get("code-package")
        env_id = json.loads(metadata.get("conda_env_id", "{}"))
        if info is None or not env_id:
            return {"type": "conda"}
        info = json.loads(info)
        _, blobdata = cls._filecache.get_data(
            info["ds_type"], flow_name, info["location"], info["sha"]
        )
        with tarfile.open(fileobj=BytesIO(blobdata), mode="r:gz") as tar:
            conda_file = tar.extractfile(CONDA_MAGIC_FILE_V2)
        if conda_file is None:
            return {"type": "conda"}
        cached_env_info = CachedEnvironmentInfo.from_dict(
            json.loads(conda_file.read().decode("utf-8"))
        )
        resolved_env = cached_env_info.env_for(env_id["req_id"], env_id["full_id"])
        if resolved_env:
            new_info = {
                "type": "conda",
                "req_id": env_id["req_id"],
                "full_id": env_id["full_id"],
                "user_deps": "; ".join(map(str, resolved_env.deps)),
                "all_packages": "; ".join([p.filename for p in resolved_env.packages]),
            }
            return new_info
        return {"type": "conda"}

    def get_package_commands(self, code_package_url: str, datastore_type: str):
        return self.base_env.get_package_commands(code_package_url, datastore_type)

    def get_environment_info(self):
        return self.base_env.get_environment_info()



from concurrent.futures import ThreadPoolExecutor, as_completed
import os
import platform

from functools import wraps
from threading import local
from typing import Any, Dict, List, Optional
from metaflow._vendor import click

from metaflow.cli import echo_dev_null, echo_always


from metaflow import Step, metaflow_config, namespace
from metaflow.plugins import DATASTORES
from metaflow.metaflow_config import DEFAULT_DATASTORE, DEFAULT_METADATA
from metaflow.plugins import METADATA_PROVIDERS
from metaflow_extensions.netflix_ext.plugins.conda.conda import Conda
from metaflow_extensions.netflix_ext.plugins.conda.env_descr import (
    EnvID,
    ResolvedEnvironment,
)
from metaflow_extensions.netflix_ext.plugins.conda.utils import arch_id

from .utils import (
    env_id_from_step,
    download_mf_version,
    local_instances_for_req_id,
    parse_channels,
    parse_deps,
    pretty_print_env,
    quiet_print_env,
    req_id_from_spec,
)


class CommandObj:
    def __init__(self):
        pass


def env_spec_options(func):
    @click.option(
        "--python",
        default=platform.python_version(),
        show_default=True,
        help="Python version for the environment",
    )
    @click.option(
        "--deps",
        required=True,
        help="Semi-colon separated list of dependencies. Specify version constraints per "
        "dependency using a comma separated list of constraints",
    )
    @click.option(
        "--channels",
        default="",
        show_default=True,
        help="Comma separated list of channels to search in addition to the default ones",
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


@click.group()
@click.pass_context
def cli(ctx):
    pass


@cli.group(help="Execution environment related commands.")
@click.option(
    "--quiet/--not-quiet",
    show_default=True,
    default=False,
    help="Suppress unnecessary messages",
)
@click.option(
    "--metadata",
    default=DEFAULT_METADATA,
    show_default=True,
    type=click.Choice([m.TYPE for m in METADATA_PROVIDERS]),
    help="Metadata service type",
)
@click.option(
    "--datastore",
    default=DEFAULT_DATASTORE,
    show_default=True,
    type=click.Choice([d.TYPE for d in DATASTORES]),
    help="Data backend type",
)
# For later support of different environment types.
@click.option(
    "--environment",
    default="conda",
    show_default=True,
    type=click.Choice(["conda"]),
    help="Type of environment to manage",
)
@click.option(
    "--conda-root", default=None, help="Root path for Conda cached information"
)
@click.pass_context
def environment(
    ctx: Any,
    quiet: bool,
    metadata: str,
    datastore: str,
    environment: str,
    conda_root: Optional[str],
):
    global echo
    if quiet:
        echo = echo_dev_null
    else:
        echo = echo_always

    obj = CommandObj()
    obj.quiet = quiet
    obj.echo = echo
    obj.echo_always = echo_always
    obj.datastore_type = datastore

    if conda_root:
        metaflow_config.CONDA_S3ROOT = conda_root
    obj.conda_root = metaflow_config.CONDA_S3ROOT

    obj.metadata_type = metadata

    obj.conda = Conda(obj.echo, obj.datastore_type)
    ctx.obj = obj


# @environment.command(help="List resolved environments for a set of dependencies")
# @click.option(
#     "--local-only/--no-local-only",
#     default=False,
#     show_default=True,
#     help="Only list environments locally known",
# )
# @click.option(
#     "--archs",
#     default=arch_id(),
#     show_default=True,
#     help="Comma separated list of architectures to list environments for",
# )
# @click.pass_obj
# @env_spec_options
# def list(obj, local_only, archs, python, deps, channels):
#     req_id = req_id_from_spec(python, deps, channels)
#     my_arch = arch_id()

#     obj.echo(
#         "Listing environments for python: %s, dependencies: %s and channels: %s "
#         "(requirement hash: %s)"
#         % (python, str(parse_deps(deps)), str(parse_channels(channels)), req_id)
#     )

#     # Get all the environments that we know about
#     envs = []
#     for arch in archs.split(","):
#         envs.extend(obj.conda.environments(req_id, arch, local_only))

#     # Get the local environments so we can say if an environment is present
#     local_instances = local_instances_for_req_id(obj.conda, req_id)

#     if obj.quiet:
#         obj.echo_always(
#             "# req_id full_id arch resolved_on resolved_by resolved_on packages local_instances"
#         )
#         for env_id, env in envs:
#             quiet_print_env(
#                 obj, env, local_instances=local_instances.get(env_id.full_id)
#             )
#     else:
#         # Extract the mapping between default and regular env
#         default_env_id = {}
#         for env_id, env in envs:
#             if env_id.full_id == "_default":
#                 default_env_id[env_id.arch] = env.env_id

#         # Print out the environments
#         for env_id, env in envs:
#             if env_id.full_id == "_default":
#                 # We will print it out as a regular environment
#                 continue
#             pretty_print_env(
#                 obj,
#                 env,
#                 env.env_id == default_env_id.get(env_id.arch),
#                 local_instances=local_instances.get(env_id.full_id),
#             )
#         if len(envs) == 0:
#             if local_only:
#                 obj.echo(
#                     "*No environments exist locally, try without "
#                     "--local-only to search remotely*"
#                 )
#             else:
#                 obj.echo("*No environments exist*")
#         if default_env_id is None:
#             obj.echo("\n\nNo default environment defined -- will resolve when needed")


# @environment.command(help="Resolve an environment")
# @click.option(
#     "--archs",
#     default=arch_id(),
#     show_default=True,
#     help="Co-resolve for all the architectures listed (comma-separated list)",
# )
# @click.option(
#     "--force/--no-force",
#     default=False,
#     show_default=True,
#     help="Force re-resolution even if already resolved",
# )
# @click.option(
#     "--cache/--no-cache",
#     default=True,
#     show_default=True,
#     help="Cache resolved environment and packages if possible",
# )
# @click.pass_obj
# @env_spec_options
# def resolve(obj, archs, force, cache, python, deps, channels):
#     def _inner_resolve(
#         deps: List[str], channels: List[str], arch: str
#     ) -> ResolvedEnvironment:
#         resolved_env = obj.conda.resolve(["<local>"], deps, channels, arch)
#         return resolved_env

#     req_id = req_id_from_spec(python, deps, channels)
#     deps = parse_deps(deps) + ["python==%s" % python]
#     channels = parse_channels(channels)
#     archs = archs.split(",")
#     my_arch = arch_id()
#     if obj.conda.environment(EnvID(req_id=req_id, full_id="_default", arch=my_arch)):
#         if force:
#             obj.echo("Re-resolving environment")
#         else:
#             raise ValueError(
#                 "Environment already exists -- use --force to force resolution"
#             )
#     envs = []  # type: List[ResolvedEnvironment]
#     obj.echo(
#         "Resolving for dependencies: %s and channels: %s on %s (requirement hash: %s)..."
#         % (str(deps), str(channels), str(archs), req_id)
#     )
#     with ThreadPoolExecutor() as executor:
#         resolution_result = [
#             executor.submit(_inner_resolve, deps, channels, arch) for arch in archs
#         ]
#         for f in as_completed(resolution_result):
#             envs.append(f.result())
#     ResolvedEnvironment.set_coresolved_full_id(envs)

#     if cache:
#         obj.echo("Caching environments...")
#         obj.conda.cache_environments(envs)

#     obj.conda.add_environments(envs)
#     # We always resolve for the "_default" environment so update that
#     for env in envs:
#         obj.conda.set_default_environment(env.env_id)
#     obj.conda.write_out_environments()

#     # Get the local environments so we can say if an environment is present
#     local_instances = local_instances_for_req_id(obj.conda, req_id)

#     if obj.quiet:
#         for env in envs:
#             quiet_print_env(
#                 obj, env, local_instances=local_instances.get(env.env_id.full_id)
#             )
#     else:
#         for env in envs:
#             pretty_print_env(
#                 obj,
#                 env,
#                 is_default=True,
#                 local_instances=local_instances.get(env.env_id.full_id),
#             )


@environment.command()
@click.argument(
    "envspec",
    nargs=-1,
)
@click.option("--name", default=None, help="Name of the environment to create")
@click.option(
    "--recreate/--no-recreate",
    default=False,
    show_default=True,
    help="Recreate the environment if it already exists",
)
@click.option(
    "--into",
    default=None,
    show_default=False,
    type=click.Path(file_okay=False, writable=True, readable=True, resolve_path=True),
    help="If specified and for pathspec ENVSPEC, downloads the code to this directory",
)
@click.pass_obj
def create_local(obj, name, recreate, into, envspec):
    """
    Create a local environment based on ENVSPEC.

    ENVSPEC can either be:
      - a pathspec in the form <flowname>/<runid>/<stepname>
      - two values: the requirement hash and the full hash of the environment.

    The latter two values are returned for example when using `list` for example
    """
    if into:
        if os.path.exists(into):
            if len(os.listdir(into)) != 0:
                raise ValueError("Directory '%s' is not empty" % into)
        else:
            os.makedirs(into)
    code_pkg = None
    if len(envspec) == 1:
        # This should be a pathspec
        if len(envspec[0].split("/")) != 3:
            raise ValueError(
                "Pathspec should be in the form <flowname>/<runid>/<stepname>"
            )
        namespace(None)
        step = Step(
            envspec[0]
        )  # Will raise MetaflowNotFound as expected if it doesn't exist
        env_id = env_id_from_step(step)
        code_pkg = step.task.code
        mf_version = step.task.metadata_dict["metaflow_version"]
    elif len(envspec) == 2:
        # This should be req_id and full_id
        req_id, full_id = envspec
        env_id = EnvID(req_id=req_id, full_id=full_id, arch=arch_id())
        if into:
            obj.echo("Ignoring --into option since a pathspec was not given")
            into = None
    else:
        raise ValueError("envspec not in the expected format")

    env = obj.conda.environment(env_id, local_only=True)
    if env is None:
        env = obj.conda.environment(env_id)
        # We are going to cache locally as well
        if env:
            obj.conda.write_out_environments()
    if env is None:
        raise ValueError(
            "Environment for requirement hash %s (full hash %s) does not exist for "
            "this architecture" % (env_id.req_id, env_id.full_id)
        )
    name = name or "metaflowtmp_%s_%s" % (env_id.req_id, env_id.full_id)

    existing_env = obj.conda.created_environment(name)
    if existing_env:
        if not recreate:
            raise ValueError(
                "Environment '%s' already exists; use --recreate to force recreating"
                % name
            )
        obj.conda.remove_for_name(name)

    if into:
        os.chdir(into)
        obj.conda.create_for_name(name, env, do_symlink=True)
        if code_pkg:
            code_pkg.tarball.extractall(path=".")
        else:
            # We get metaflow version and recreate it in the directory
            obj.echo(
                "Step '%s' does not have a code package -- "
                "downloading active Metaflow version only" % envspec[0]
            )
            download_mf_version("./__conda_python", mf_version)
        obj.echo(
            "Code package for %s downloaded into '%s' -- `__conda_python` is "
            "the executable to use" % (envspec[0], into)
        )
    else:
        obj.conda.create_for_name(name, env)

    obj.echo(
        "Created environment '%s' locally, activate with `%s activate %s`"
        % (name, obj.conda.binary("conda"), name)
    )



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false

import importlib
import json
import os
import platform
import shutil
import sys
import tempfile

from itertools import chain

from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    cast,
)

from metaflow.plugins.datastores.local_storage import LocalStorage
from metaflow.datastore.flow_datastore import FlowDataStore
from metaflow.datastore.task_datastore import TaskDataStore
from metaflow.debug import debug
from metaflow.decorators import Decorator, StepDecorator
from metaflow.extension_support import EXT_PKG
from metaflow.flowspec import FlowSpec
from metaflow.graph import FlowGraph
from metaflow.metadata import MetaDatum
from metaflow.metadata.metadata import MetadataProvider
from metaflow.metaflow_config import (
    CONDA_REMOTE_COMMANDS,
    CONDA_PREFERRED_FORMAT,
    get_pinned_conda_libs,
)
from metaflow.metaflow_environment import (
    InvalidEnvironmentException,
    MetaflowEnvironment,
)
from .env_descr import (
    TStr,
    EnvID,
    ResolvedEnvironment,
)
from metaflow.plugins.env_escape import generate_trampolines
from metaflow.unbounded_foreach import UBF_CONTROL
from metaflow.util import get_metaflow_root

from .utils import arch_id
from .conda import Conda


class CondaStepDecorator(StepDecorator):
    """
    Specifies the Conda environment for the step.

    Information in this decorator will augment any
    attributes set in the `@conda_base` flow-level decorator. Hence
    you can use `@conda_base` to set common libraries required by all
    steps and use `@conda` to specify step-specific additions.

    Parameters
    ----------
    libraries : Dict
        Libraries to use for this step. The key is the name of the package
        and the value is the version to use (default: `{}`).
    python : string
        Version of Python to use, e.g. '3.7.4'
        (default: None, i.e. the current Python version).
    disabled : bool
        If set to True, disables Conda (default: False).
    """

    name = "conda"
    defaults = {
        "libraries": {},
        "channels": [],
        "pip_packages": {},
        "pip_sources": [],
        "archs": None,
        "python": None,
        "disabled": None,
    }  # type: Dict[str, Any]

    conda = None  # type: Optional[Conda]
    _local_root = None  # type: Optional[str]

    def is_enabled(self, ubf_context: Optional[str] = None) -> bool:
        if ubf_context == UBF_CONTROL:
            return False
        return not next(
            x
            for x in [
                self.attributes["disabled"],
                self._base_attributes["disabled"],
                False,
            ]
            if x is not None
        )

    @property
    def env_id(self) -> EnvID:
        return EnvID(
            req_id=ResolvedEnvironment.get_req_id(self.step_deps, self.source_deps),
            full_id="_default",
            arch=arch_id(),
        )

    @property
    def source_deps(self) -> Sequence[TStr]:
        sources = list(
            map(
                lambda x: TStr("conda", x),
                self._conda_channels(),
            )
        )

        sources.extend(
            map(
                lambda x: TStr("pip", x),
                self._pip_sources(),
            )
        )
        return sources

    @property
    def step_deps(self) -> Sequence[TStr]:
        deps = [TStr("conda", "python==%s" % self._python_version())]
        deps.extend(
            TStr("conda", "%s==%s" % (name, ver))
            for name, ver in self._conda_deps().items()
        )
        deps.extend(
            TStr("pip", "%s==%s" % (name, ver))
            for name, ver in self._pip_deps().items()
        )
        return deps

    @property
    def requested_architectures(self) -> List[str]:
        return self._archs

    @property
    def local_root(self) -> Optional[str]:
        return self._local_root

    def step_init(
        self,
        flow: FlowSpec,
        graph: FlowGraph,
        step_name: str,
        decorators: Sequence[StepDecorator],
        environment: MetaflowEnvironment,
        flow_datastore: FlowDataStore,
        logger: Callable[..., None],
    ):
        if environment.TYPE != "conda":
            raise InvalidEnvironmentException(
                "The *@conda* decorator requires " "--environment=conda"
            )

        self._echo = logger
        self._env = environment
        self._flow = flow
        self._step_name = step_name
        self._flow_datastore_type = flow_datastore.TYPE  # type: str
        self._base_attributes = self._get_base_attributes()

        self._archs = self._architectures(decorators)

        self.__class__._local_root = LocalStorage.get_datastore_root_from_config(
            self._echo
        )  # type: str
        os.environ["PYTHONNOUSERSITE"] = "1"

    def runtime_init(self, flow: FlowSpec, graph: FlowGraph, package: Any, run_id: str):
        # Create a symlink to installed version of metaflow to execute user code against
        path_to_metaflow = os.path.join(get_metaflow_root(), "metaflow")
        path_to_info = os.path.join(get_metaflow_root(), "INFO")
        self._metaflow_home = tempfile.mkdtemp(dir="/tmp")
        self.addl_paths = None  # type: Optional[List[str]]
        os.symlink(path_to_metaflow, os.path.join(self._metaflow_home, "metaflow"))

        # Symlink the INFO file as well to properly propagate down the Metaflow version
        # if launching on AWS Batch for example
        if os.path.isfile(path_to_info):
            os.symlink(path_to_info, os.path.join(self._metaflow_home, "INFO"))
        else:
            # If there is no "INFO" file, we will actually create one in this new
            # place because we won't be able to properly resolve the EXT_PKG extensions
            # the same way as outside conda (looking at distributions, etc). In a
            # Conda environment, as shown below (where we set self.addl_paths), all
            # EXT_PKG extensions are PYTHONPATH extensions. Instead of re-resolving,
            # we use the resolved information that is written out to the INFO file.
            with open(
                os.path.join(self._metaflow_home, "INFO"), mode="wt", encoding="utf-8"
            ) as f:
                f.write(json.dumps(self._env.get_environment_info()))

        # Do the same for EXT_PKG
        try:
            m = importlib.import_module(EXT_PKG)
        except ImportError:
            # No additional check needed because if we are here, we already checked
            # for other issues when loading at the toplevel
            pass
        else:
            custom_paths = list(set(m.__path__))  # For some reason, at times, unique
            # paths appear multiple times. We simplify
            # to avoid un-necessary links

            if len(custom_paths) == 1:
                # Regular package; we take a quick shortcut here
                os.symlink(
                    custom_paths[0],
                    os.path.join(self._metaflow_home, EXT_PKG),
                )
            else:
                # This is a namespace package, we therefore create a bunch of directories
                # so we can symlink in those separately and we will add those paths
                # to the PYTHONPATH for the interpreter. Note that we don't symlink
                # to the parent of the package because that could end up including
                # more stuff we don't want
                self.addl_paths = []
                for p in custom_paths:
                    temp_dir = tempfile.mkdtemp(dir=self._metaflow_home)
                    os.symlink(p, os.path.join(temp_dir, EXT_PKG))
                    self.addl_paths.append(temp_dir)

        # Also install any environment escape overrides directly here to enable
        # the escape to work even in non metaflow-created subprocesses
        generate_trampolines(self._metaflow_home)

    def runtime_step_cli(
        self,
        cli_args: Any,  # Importing CLIArgs causes an issue so ignore for now
        retry_count: int,
        max_user_code_retries: int,
        ubf_context: str,
    ):
        no_remote = all([x not in cli_args.commands for x in CONDA_REMOTE_COMMANDS])
        if self.is_enabled(ubf_context) and no_remote:
            self._get_conda(self._echo, self._flow_datastore_type)
            assert self.conda
            resolved_env = cast(
                ResolvedEnvironment, self.conda.environment(self.env_id)
            )
            my_env_id = resolved_env.env_id
            # Export this for local runs, we will use it to read the "resolved"
            # environment ID in task_pre_step; this makes it compatible with the remote
            # bootstrap which also exports it.
            cli_args.env["_METAFLOW_CONDA_ENV"] = json.dumps(my_env_id)
            # Create the environment we are going to use
            if self.conda.created_environment(my_env_id):
                self._echo(
                    "Using existing Conda environment %s (%s)"
                    % (my_env_id.req_id, my_env_id.full_id)
                )
            else:
                # Otherwise, we read the conda file and create the environment locally
                self._echo(
                    "Creating Conda environment %s (%s)..."
                    % (my_env_id.req_id, my_env_id.full_id)
                )
                self.conda.create_for_step(self._step_name, resolved_env)

            # Actually set it up.
            python_path = self._metaflow_home
            if self.addl_paths is not None:
                addl_paths = os.pathsep.join(self.addl_paths)
                python_path = os.pathsep.join([addl_paths, python_path])

            cli_args.env["PYTHONPATH"] = python_path
            cli_args.entrypoint[0] = self.conda.python(my_env_id)

    def task_pre_step(
        self,
        step_name: str,
        task_datastore: TaskDataStore,
        metadata: MetadataProvider,
        run_id: str,
        task_id: str,
        flow: FlowSpec,
        graph: FlowGraph,
        retry_count: int,
        max_user_code_retries: int,
        ubf_context: str,
        inputs: List[str],
    ):
        if self.is_enabled(ubf_context):
            # Add the Python interpreter's parent to the path. This is to
            # ensure that any non-pythonic dependencies introduced by the conda
            # environment are visible to the user code.
            env_path = os.path.dirname(os.path.realpath(sys.executable))
            if os.environ.get("PATH") is not None:
                env_path = os.pathsep.join([env_path, os.environ["PATH"]])
            os.environ["PATH"] = env_path

            metadata.register_metadata(
                run_id,
                step_name,
                task_id,
                [
                    MetaDatum(
                        field="conda_env_id",
                        value=os.environ["_METAFLOW_CONDA_ENV"],
                        type="conda_env_id",
                        tags=["attempt_id:{0}".format(retry_count)],
                    )
                ],
            )

    def runtime_finished(self, exception: Exception):
        shutil.rmtree(self._metaflow_home)

    def _get_base_attributes(self) -> Dict[str, Any]:
        if "conda_base" in self._flow._flow_decorators:
            return self._flow._flow_decorators["conda_base"].attributes
        return self.defaults

    def _python_version(self) -> str:
        return next(
            x
            for x in [
                self.attributes["python"],
                self._base_attributes["python"],
                platform.python_version(),
            ]
            if x is not None
        )

    def _conda_deps(self) -> Dict[str, str]:
        deps = get_pinned_conda_libs(self._python_version(), self._flow_datastore_type)

        deps.update(self._base_attributes["libraries"])
        deps.update(self.attributes["libraries"])

        return deps

    def _conda_channels(self) -> List[str]:
        return list(
            chain(self.attributes["channels"], self._base_attributes["channels"])
        )

    def _pip_deps(self) -> Dict[str, str]:
        # TODO: Do we want to add pinned pip libs?
        deps = {}  # type: Dict[str, str]
        deps.update(self._base_attributes["pip_packages"])
        deps.update(self.attributes["pip_packages"])

        return deps

    def _pip_sources(self) -> List[str]:
        return list(
            chain(self.attributes["pip_sources"], self._base_attributes["pip_sources"])
        )

    def _architectures(self, decos: Iterable[Decorator]) -> List[str]:
        archs = [arch_id()]
        for deco in decos:
            if deco.name in CONDA_REMOTE_COMMANDS:
                # force conda resolution for linux-64 architectures
                archs.append("linux-64")
        archs.extend(
            next(
                x
                for x in [self.attributes["archs"], self._base_attributes["archs"], []]
                if x is not None
            )
        )
        return list(set(archs))

    @classmethod
    def _get_conda(cls, echo: Callable[..., None], datastore_type: str) -> None:
        if cls.conda is None:
            cls.conda = Conda(echo, datastore_type)


def get_conda_decorator(flow: FlowSpec, step_name: str) -> CondaStepDecorator:
    step = next(step for step in flow if step.name == step_name)
    decorator = next(
        deco for deco in step.decorators if isinstance(deco, CondaStepDecorator)
    )
    # Guaranteed to have a conda decorator because of self.decospecs()
    return decorator



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false
from datetime import datetime
import errno
import fcntl
import json
import os
from hashlib import md5, sha1, sha256
from itertools import chain
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)
from urllib.parse import urlparse

from metaflow.metaflow_config import ENV_PACKAGES_DIRNAME
from metaflow.util import get_username

from .utils import (
    CONDA_FORMATS,
    TRANSMUT_PATHCOMPONENT,
    CondaException,
    arch_id,
    convert_filepath,
    get_conda_manifest_path,
)

# Order should be maintained
EnvID = NamedTuple("EnvID", [("req_id", str), ("full_id", str), ("arch", str)])


class TStr:
    def __init__(self, category: str, value: str):
        self._category = category
        self._value = value

    @property
    def category(self):
        return self._category

    @property
    def value(self):
        return self._value

    @value.setter
    def value_setter(self, value: str):
        self._value = value

    def __str__(self):
        return "%s::%s" % (self._category, self._value)

    def __repr__(self):
        return str(self)

    @staticmethod
    def from_str(value: str):
        splits = value.split("::", 1)
        if len(splits) != 2:
            raise ValueError("Cannot parse a TStr from %s" % value)
        return TStr(splits[0], splits[1])


class CachePackage:

    TYPE = "invalid"

    _class_per_type = None  # type: Optional[Dict[str, Type[CachePackage]]]

    @staticmethod
    def _ensure_class_per_type():
        if CachePackage._class_per_type is None:
            CachePackage._class_per_type = {
                c.TYPE: c for c in CachePackage.__subclasses__()
            }

    @classmethod
    def make_cache_url(
        cls,
        base_url: str,
        file_hash: str,
        file_format: Optional[str] = None,
        is_transmuted: bool = False,
    ) -> str:
        cls._ensure_class_per_type()
        url = urlparse(base_url)
        file_path, filename = convert_filepath(url.path, file_format)

        if is_transmuted:
            return os.path.join(
                cast(str, ENV_PACKAGES_DIRNAME),
                cls.TYPE,
                TRANSMUT_PATHCOMPONENT,
                url.netloc,
                file_path.lstrip("/"),
                filename,
                file_hash,
                filename,
            )
        else:
            return os.path.join(
                cast(str, ENV_PACKAGES_DIRNAME),
                cls.TYPE,
                url.netloc,
                file_path.lstrip("/"),
                filename,
                file_hash,
                filename,
            )

    def __init__(self, url: str):

        self._url = url
        basename, filename = os.path.split(url)

        self._pkg_fmt = None
        for f in self.allowed_formats():
            if filename.endswith(f):
                self._pkg_fmt = f
                break
        else:
            raise ValueError(
                "URL '%s' does not end with a supported file format %s"
                % (url, str(self.allowed_formats()))
            )
        basename, self._hash = os.path.split(basename)
        self._is_transmuted = TRANSMUT_PATHCOMPONENT in basename

    @classmethod
    def allowed_formats(cls) -> Sequence[str]:
        raise NotImplementedError()

    @property
    def url(self) -> str:
        return self._url

    @property
    def hash(self) -> str:
        return self._hash

    @property
    def format(self) -> str:
        return self._pkg_fmt  # type: ignore

    @property
    def is_transmuted(self) -> bool:
        return self._is_transmuted

    def to_dict(self) -> Dict[str, Any]:
        return {"_type": self.TYPE, "url": self._url}

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]):
        cls._ensure_class_per_type()
        assert cls._class_per_type
        return cls._class_per_type[d["_type"]](url=d["url"])

    def __str__(self):
        return "%s#%s" % (self.url, self.hash)


class CondaCachePackage(CachePackage):
    TYPE = "conda"

    @classmethod
    def allowed_formats(cls) -> Sequence[str]:
        from .utils import CONDA_FORMATS

        return CONDA_FORMATS


class PipCachePackage(CachePackage):
    TYPE = "pip"

    @classmethod
    def allowed_formats(cls) -> Sequence[str]:
        return [".whl", ".tar.gz"]


class PackageSpecification:
    TYPE = "invalid"

    _class_per_type = None  # type: Optional[Dict[str, Type[PackageSpecification]]]

    def __init__(
        self,
        filename: str,
        url: str,
        url_format: Optional[str] = None,
        hashes: Optional[Dict[str, str]] = None,
        cache_info: Optional[Dict[str, CachePackage]] = None,
    ):
        # if "/" in filename and not filename.startswith(self.TYPE):
        #     raise ValueError(
        #         "Attempting to create a package of type %s with filename %s"
        #         % (self.TYPE, filename)
        #     )
        # if self.TYPE != "conda" and not filename.startswith("%s/" % self.TYPE):  # type: ignore
        #     filename = "/".join([self.TYPE, filename])
        self._filename = filename
        (
            self._package_name,
            self._package_version,
            self._package_detailed_version,
        ) = self._split_filename()

        self._url = url
        if url_format is None:
            for ending in self.allowed_formats():
                if self._url.endswith(ending):
                    url_format = ending
                    break
            else:
                raise ValueError(
                    "URL '%s' does not end in a known ending (%s)"
                    % (self._url, str(self.allowed_formats()))
                )
        self._url_format = url_format
        self._hashes = hashes or {}
        self._cache_info = cache_info or {}

        # Additional information used for local book-keeping as we are updating
        # the package
        self._local_dir = None  # type: Optional[str]
        self._local_path = {}  # type: Dict[str, str]
        self._is_fetched = []  # type: List[str]
        self._is_transmuted = []  # type: List[str]

    @property
    def filename(self) -> str:
        return self._filename

    @property
    def package_name(self) -> str:
        return self._package_name

    @property
    def package_version(self) -> str:
        return self._package_version

    @property
    def package_detailed_version(self) -> str:
        return self._package_detailed_version

    @property
    def url(self) -> str:
        return self._url

    @property
    def url_format(self) -> str:
        return self._url_format

    @classmethod
    def cache_pkg_type(cls) -> Type[CachePackage]:
        raise NotImplementedError

    def pkg_hash(self, pkg_format: str) -> Optional[str]:
        return self._hashes.get(pkg_format)

    @property
    def pkg_hashes(self) -> Iterable[Tuple[str, str]]:
        for pkg_fmt, pkg_hash in self._hashes.items():
            yield (pkg_fmt, pkg_hash)

    def add_pkg_hash(self, pkg_format: str, pkg_hash: str):
        old_hash = self.pkg_hash(pkg_format)
        if old_hash and old_hash != pkg_hash:
            raise ValueError(
                "Attempting to add an inconsistent hash for package %s; "
                "adding %s when already have %s" % (self.filename, pkg_hash, old_hash)
            )
        self._hashes[pkg_format] = pkg_hash

    @property
    def local_dir(self) -> Optional[str]:
        # Returns the local directory found for this package
        return self._local_dir

    def local_file(self, pkg_format: str) -> Optional[str]:
        # Return the local tar-ball for this package (depending on the format)
        return self._local_path.get(pkg_format)

    @property
    def local_files(self) -> Iterable[Tuple[str, str]]:
        for pkg_fmt, local_path in self._local_path.items():
            yield (pkg_fmt, local_path)

    def is_fetched(self, pkg_format: str) -> bool:
        # Return whether the local tar-ball for this package had to be fetched from
        # either cache or web
        return pkg_format in self._is_fetched

    def is_transmuted(self, pkg_format: str) -> bool:
        return pkg_format in self._is_transmuted

    def add_local_dir(self, local_path: str):
        # Add a local directory that is present for this package
        local_dir = self.local_dir
        if local_dir and local_dir != local_path:
            raise ValueError(
                "Attempting to add an inconsistent local directory for package %s; "
                "adding %s when already have %s"
                % (self.filename, local_path, local_dir)
            )
        self._local_dir = local_path

    def add_local_file(
        self,
        pkg_format: str,
        local_path: str,
        pkg_hash: Optional[str] = None,
        downloaded: bool = False,
        transmuted: bool = False,
    ):
        # Add a local file for this package indicating whether it was downloaded or
        # transmuted
        existing_path = self.local_file(pkg_format)
        if existing_path and local_path != existing_path:
            raise ValueError(
                "Attempting to add inconsistent local files of format %s for a package %s; "
                "adding %s when already have %s"
                % (pkg_format, self.filename, local_path, existing_path)
            )
        self._local_path[pkg_format] = local_path
        known_hash = self._hashes.get(pkg_format)
        added_hash = pkg_hash or self._hash_pkg(local_path)
        if known_hash and known_hash != added_hash:
            raise ValueError(
                "Attempting to add inconsistent local files of format %s for package %s; "
                "got a hash of %s but expected %s"
                % (pkg_format, self.filename, added_hash, known_hash)
            )
        self._hashes[pkg_format] = added_hash
        if downloaded:
            self._is_fetched.append(pkg_format)
        if transmuted:
            self._is_transmuted.append(pkg_format)

    def cached_version(self, pkg_format: str) -> Optional[CachePackage]:
        return self._cache_info.get(pkg_format)

    @property
    def cached_versions(self) -> Iterable[Tuple[str, CachePackage]]:
        for pkg_fmt, cached in self._cache_info.items():
            yield (pkg_fmt, cached)

    def add_cached_version(self, pkg_format: str, cache_info: CachePackage) -> None:
        if cache_info.TYPE != self.TYPE:
            raise ValueError(
                "Attempting to add a cache package of type %s for a package of type %s"
                % (cache_info.TYPE, self.TYPE)
            )

        old_cache_info = self.cached_version(pkg_format)
        if old_cache_info and (
            old_cache_info.url != cache_info.url
            or old_cache_info.hash != cache_info.hash
        ):
            raise ValueError(
                "Attempting to add inconsistent cache information for format %s of package %s; "
                "adding %s when already have %s"
                % (pkg_format, self.filename, cache_info, old_cache_info)
            )

        old_pkg_hash = self.pkg_hash(pkg_format)
        if old_pkg_hash and old_pkg_hash != cache_info.hash:
            raise ValueError(
                "Attempting to add inconsistent cache information for format %s of package %s; "
                "adding a package with hash %s when expected %s"
                % (pkg_format, self.filename, cache_info.hash, old_pkg_hash)
            )

        self._cache_info[pkg_format] = cache_info
        self._hashes[pkg_format] = cache_info.hash

    def is_cached(self, formats: List[str]) -> bool:
        if formats:
            return all([f in self._cache_info for f in formats])
        return len(self._cache_info) > 0

    def to_dict(self) -> Dict[str, Any]:
        d = {
            "_type": self.TYPE,
            "filename": self.filename,
            "url": self.url,
            "url_format": self.url_format,
            "hashes": self._hashes,
        }  # type: Dict[str, Any]
        if self._cache_info:
            cache_d = {
                pkg_format: info.to_dict()
                for pkg_format, info in self._cache_info.items()
            }
            d["cache_info"] = cache_d
        return d

    def _split_filename(self) -> Tuple[str, str, str]:
        raise NotImplementedError

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]):
        if cls._class_per_type is None:
            cls._class_per_type = {c.TYPE: c for c in cls.__subclasses__()}

        cache_info = d.get("cache_info", {})  # type: Dict[str, Any]
        url_format = d.get("url_format")
        return cls._class_per_type[d["_type"]](
            filename=d["filename"],
            url=d["url"],
            url_format=url_format,
            hashes=d["hashes"],
            cache_info={
                pkg_fmt: CachePackage.from_dict(info)
                for pkg_fmt, info in cache_info.items()
            },
        )

    @classmethod
    def allowed_formats(cls) -> Sequence[str]:
        raise NotImplementedError()

    @classmethod
    def base_hash(cls):
        raise NotImplementedError()

    @classmethod
    def _hash_pkg(cls, path: str) -> str:
        base_hash = cls.base_hash()
        with open(path, "rb") as f:
            for byte_block in iter(lambda: f.read(8192), b""):
                base_hash.update(byte_block)
        return base_hash.hexdigest()


class CondaPackageSpecification(PackageSpecification):
    TYPE = "conda"

    @classmethod
    def cache_pkg_type(cls):
        return CondaCachePackage

    @classmethod
    def allowed_formats(cls) -> Sequence[str]:
        from .utils import CONDA_FORMATS

        return CONDA_FORMATS

    @classmethod
    def base_hash(cls):
        return md5()

    def _split_filename(self) -> Tuple[str, str, str]:
        pkg, v, addl = self._filename.rsplit("-", 2)
        return pkg, v, "-".join([v, addl])


class PipPackageSpecification(PackageSpecification):
    TYPE = "pip"

    @classmethod
    def cache_pkg_type(cls):
        return PipCachePackage

    @classmethod
    def allowed_formats(cls) -> Sequence[str]:
        return [".whl", ".tar.gz"]

    @classmethod
    def base_hash(cls):
        return sha256()

    def _split_filename(self) -> Tuple[str, str, str]:
        name, version, _, _, _ = self._filename.rsplit("-", 4)
        if "." in name:
            # This means the version is actually a build identifier and version
            # is in name
            name, version = name.rsplit("-", 1)
        return (name, version, version)


TResolvedEnvironment = TypeVar("TResolvedEnvironment", bound="ResolvedEnvironment")


class ResolvedEnvironment:
    def __init__(
        self,
        user_dependencies: Sequence[TStr],
        user_sources: Optional[Sequence[TStr]],
        arch: Optional[str] = None,
        env_id: Optional[EnvID] = None,
        user_alias: Optional[str] = None,
        all_packages: Optional[Sequence[PackageSpecification]] = None,
        resolved_on: Optional[datetime] = None,
        resolved_by: Optional[str] = None,
        co_resolved: Optional[List[str]] = None,
    ):
        self._user_dependencies = list(user_dependencies)
        self._user_sources = list(user_sources) if user_sources else []
        if not env_id:
            env_req_id = ResolvedEnvironment.get_req_id(
                self._user_dependencies, self._user_sources
            )
            env_full_id = "_unresolved"
            if all_packages is not None:
                env_full_id = self._compute_hash([p.filename for p in all_packages])
            self._env_id = EnvID(
                req_id=env_req_id, full_id=env_full_id, arch=arch or arch_id()
            )
        else:
            self._env_id = env_id
        self._alias = user_alias
        self._all_packages = list(all_packages) if all_packages else []
        self._resolved_on = resolved_on or datetime.now()
        self._resolved_by = resolved_by or get_username() or "unknown"
        self._co_resolved = co_resolved or [arch or arch_id()]

    @staticmethod
    def get_req_id(
        deps: Sequence[TStr], sources: Optional[Sequence[TStr]] = None
    ) -> str:
        # Extract per category so we can sort independently for each category
        deps_by_category = {}  # type: Dict[str, List[str]]
        sources_by_category = {}  # type: Dict[str, List[str]]
        for d in deps:
            deps_by_category.setdefault(d.category, []).append(d.value)
        if sources:
            for s in sources:
                sources_by_category.setdefault(s.category, []).append(s.value)
        return ResolvedEnvironment._compute_hash(
            chain(
                *(sorted(deps_by_category[c]) for c in sorted(deps_by_category)),
                *(sorted(sources_by_category[c]) for c in sorted(sources_by_category))
            )
        )

    @staticmethod
    def set_coresolved_full_id(envs: Sequence[TResolvedEnvironment]) -> None:
        envs = sorted(envs, key=lambda x: x.env_id.arch)
        to_hash = []  # type: List[str]
        archs = []  # type: List[str]
        for env in envs:
            archs.append(env.env_id.arch)
            to_hash.append(env.env_id.arch)
            to_hash.extend([p.filename for p in env.packages])
        new_full_id = ResolvedEnvironment._compute_hash(to_hash)
        for env in envs:
            env.set_coresolved(archs, new_full_id)

    @property
    def deps(self) -> List[TStr]:
        return self._user_dependencies

    @property
    def sources(self) -> List[TStr]:
        return self._user_sources

    @property
    def env_id(self) -> EnvID:
        if self._env_id.full_id in ("_default", "_unresolved") and self._all_packages:
            env_full_id = self._compute_hash([p.filename for p in self._all_packages])
            self._env_id = self._env_id._replace(full_id=env_full_id)
        return self._env_id

    @property
    def env_alias(self) -> Optional[str]:
        return self._alias

    @property
    def packages(self) -> Iterable[PackageSpecification]:
        for p in self._all_packages:
            yield p

    @property
    def resolved_on(self) -> datetime:
        return self._resolved_on

    @property
    def resolved_by(self) -> str:
        return self._resolved_by

    @property
    def co_resolved_archs(self) -> List[str]:
        return self._co_resolved

    def add_package(self, pkg: PackageSpecification):
        self._all_packages.append(pkg)
        if self._env_id.full_id not in ("_default", "_unresolved"):
            self._env_id._replace(full_id="_unresolved")

    def set_coresolved(self, archs: List[str], full_id: str) -> None:
        self._env_id = EnvID(
            req_id=self._env_id.req_id, full_id=full_id, arch=self._env_id.arch
        )
        self._co_resolved = archs

    def is_cached(self, formats: Dict[str, List[str]]) -> bool:
        return all([pkg.is_cached(formats.get(pkg.TYPE, [])) for pkg in self.packages])

    def to_dict(self) -> Dict[str, Any]:
        return {
            "deps": [str(x) for x in self._user_dependencies],
            "sources": [str(x) for x in self._user_sources],
            "packages": [p.to_dict() for p in self._all_packages],
            "alias": self._alias,
            "resolved_on": self._resolved_on.isoformat(),
            "resolved_by": self._resolved_by,
            "resolved_archs": self._co_resolved,
        }

    @classmethod
    def from_dict(
        cls,
        env_id: EnvID,
        d: Mapping[str, Any],
    ):
        all_packages = [PackageSpecification.from_dict(pd) for pd in d["packages"]]
        return cls(
            user_dependencies=[TStr.from_str(x) for x in d["deps"]],
            user_sources=[TStr.from_str(x) for x in d["sources"]],
            env_id=env_id,
            user_alias=d["alias"],
            all_packages=all_packages,
            resolved_on=datetime.fromisoformat(d["resolved_on"]),
            resolved_by=d["resolved_by"],
            co_resolved=d["resolved_archs"],
        )

    @staticmethod
    def _compute_hash(inputs: Iterable[str]):
        return sha1(b" ".join([s.encode("ascii") for s in inputs])).hexdigest()


class CachedEnvironmentInfo:
    def __init__(
        self,
        step_mappings: Optional[Dict[str, Tuple[str, str]]],
        env_aliases: Optional[Dict[str, Dict[str, Tuple[str, str]]]],
        resolved_environments: Optional[
            Dict[str, Dict[str, Dict[str, Union[ResolvedEnvironment, str]]]]
        ],
    ):

        self._step_mappings = step_mappings if step_mappings else {}
        self._env_aliases = env_aliases if env_aliases else {}
        self._resolved_environments = (
            resolved_environments if resolved_environments else {}
        )

    def set_default(self, env_id: EnvID):
        per_arch_envs = self._resolved_environments.get(env_id.arch)
        if per_arch_envs is None:
            raise ValueError("No cached environments for %s" % env_id.arch)
        per_req_id_envs = per_arch_envs.get(env_id.req_id)
        if per_req_id_envs is None:
            raise ValueError(
                "No cached environments for requirement ID: %s" % env_id.req_id
            )
        per_req_id_envs["_default"] = env_id.full_id

    def get_default(self, req_id: str, arch: Optional[str]) -> Optional[EnvID]:
        arch = arch or arch_id()
        per_arch_envs = self._resolved_environments.get(arch)
        if per_arch_envs is None:
            return None
        per_req_id_envs = per_arch_envs.get(req_id)
        if per_req_id_envs is None:
            return None
        full_id = per_req_id_envs.get("_default")
        if full_id is None:
            return None
        elif isinstance(full_id, str):
            return EnvID(req_id=req_id, full_id=full_id, arch=arch)
        return full_id.env_id

    def clear_default(self, req_id: str, arch: Optional[str]) -> None:
        arch = arch or arch_id()
        per_arch_envs = self._resolved_environments.get(arch)
        if per_arch_envs is None:
            return
        per_req_id_envs = per_arch_envs.get(req_id)
        if per_req_id_envs is None:
            return
        if "_default" in per_req_id_envs:
            del per_req_id_envs["_default"]

    def add_resolved_env(self, env: ResolvedEnvironment):
        env_id = env.env_id
        per_arch_envs = self._resolved_environments.setdefault(env_id.arch, {})
        per_req_id_envs = per_arch_envs.setdefault(env_id.req_id, {})
        per_req_id_envs[env_id.full_id] = env
        if env.env_alias:
            alias_info = self._env_aliases.setdefault(env.env_alias, {})
            alias_info[env_id.arch] = (env_id.req_id, env_id.full_id)

    def env_for(
        self, req_id: str, full_id: str = "_default", arch: Optional[str] = None
    ) -> Optional[ResolvedEnvironment]:
        arch = arch or arch_id()
        per_arch_envs = self._resolved_environments.get(arch)
        if per_arch_envs:
            per_req_id_envs = per_arch_envs.get(req_id)
            if per_req_id_envs:
                if full_id == "_default":
                    full_id = per_req_id_envs.get("_default", "_invalid")  # type: ignore
                return per_req_id_envs.get(full_id)  # type: ignore
        return None

    def envs_for(
        self, req_id: str, arch: Optional[str] = None
    ) -> Iterator[Tuple[EnvID, ResolvedEnvironment]]:
        arch = arch or arch_id()
        per_arch_envs = self._resolved_environments.get(arch)
        if per_arch_envs:
            per_req_id_envs = per_arch_envs.get(req_id)
            if per_req_id_envs:
                for env in per_req_id_envs.values():
                    if isinstance(env, ResolvedEnvironment):
                        yield env.env_id, env
                    # The other case is that env is a string (when the key in
                    # per_req_id_envs is "_default" but we will list that
                    # when we get to the actual environment for it so we don't output
                    # anything.

    @property
    def envs(self) -> Iterator[Tuple[EnvID, ResolvedEnvironment]]:
        for arch, per_arch_envs in self._resolved_environments.items():
            for req_id, per_req_id_envs in per_arch_envs.items():
                for full_id, env in per_req_id_envs.items():
                    if isinstance(env, ResolvedEnvironment):
                        yield (
                            EnvID(req_id=req_id, full_id=full_id, arch=arch),
                            env,
                        )

    def to_dict(self) -> Dict[str, Any]:
        result = {
            "version": 1,
            "mappings": self._step_mappings,
            "aliases": self._env_aliases,
        }  # type: Dict[str, Any]
        resolved_envs = {}
        for arch, per_arch_envs in self._resolved_environments.items():
            per_arch_resolved_env = {}
            for req_id, per_req_id_envs in per_arch_envs.items():
                per_req_id_resolved_env = {}
                for full_id, env in per_req_id_envs.items():
                    if isinstance(env, ResolvedEnvironment):
                        per_req_id_resolved_env[full_id] = env.to_dict()
                    elif full_id == "_unresolved":
                        # This is a non resolved environment -- this should not happen
                        continue
                    else:
                        # This is for the special "_default" key
                        per_req_id_resolved_env[full_id] = env
                per_arch_resolved_env[req_id] = per_req_id_resolved_env
            resolved_envs[arch] = per_arch_resolved_env
        result["environments"] = resolved_envs
        return result

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]):
        if int(d.get("version", -1)) != 1:
            raise ValueError("Wrong version information for CachedInformationInfo")

        aliases = d.get("aliases", {})  # type: Dict[str, Dict[str, Tuple[str, str]]]
        resolved_environments_dict = d.get("environments", {})
        resolved_environments = (
            {}
        )  # type: Dict[str, Dict[str, Dict[str, Union[ResolvedEnvironment, str]]]]
        for arch, per_arch_envs in resolved_environments_dict.items():
            # Parse the aliases first to make sure we have all of them
            aliases_per_arch = aliases.get(arch, {})
            reverse_map = {v: k for k, v in aliases_per_arch.items()}
            resolved_per_req_id = {}
            for req_id, per_req_id_envs in per_arch_envs.items():
                resolved_per_full_id = (
                    {}
                )  # type: Dict[str, Union[ResolvedEnvironment, str]]
                for full_id, env in per_req_id_envs.items():
                    if full_id == "_default":
                        # Special key meaning to use this fully resolved environment
                        # for anything that matches env_id but does not have a specific
                        # mapping
                        resolved_per_full_id[full_id] = env
                    else:
                        alias = reverse_map.get((req_id, full_id), None)
                        resolved_env = ResolvedEnvironment.from_dict(
                            EnvID(req_id=req_id, full_id=full_id, arch=arch), env
                        )
                        if alias:
                            if resolved_env.env_alias != alias:
                                raise ValueError(
                                    "Alias '%s' does not match the one set in the environment '%s'"
                                    % (alias, resolved_env.env_alias)
                                )
                            del reverse_map[(req_id, full_id)]
                        resolved_per_full_id[full_id] = resolved_env
                resolved_per_req_id[req_id] = resolved_per_full_id
            resolved_environments[arch] = resolved_per_req_id
            if reverse_map:
                # If there is something left, it means we have aliases that are not resolved
                raise CondaException(
                    "Aliases %s do not map to known environments"
                    % ", ".join(d["aliases"].keys())
                )
        return cls(
            step_mappings=d["mappings"],
            env_aliases=d["aliases"],
            resolved_environments=resolved_environments,
        )


def read_conda_manifest(ds_root: str) -> CachedEnvironmentInfo:
    path = get_conda_manifest_path(ds_root)
    if os.path.exists(path) and os.path.getsize(path) > 0:
        with open(path, mode="r", encoding="utf-8") as f:
            return CachedEnvironmentInfo.from_dict(json.load(f))
    else:
        return CachedEnvironmentInfo(
            step_mappings=None, env_aliases=None, resolved_environments=None
        )


def write_to_conda_manifest(ds_root: str, info: CachedEnvironmentInfo):
    path = get_conda_manifest_path(ds_root)
    try:
        os.makedirs(os.path.dirname(path))
    except OSError as x:
        if x.errno != errno.EEXIST:
            raise
    with os.fdopen(
        os.open(path, os.O_WRONLY | os.O_CREAT | os.O_TRUNC), "w", encoding="utf-8"
    ) as f:
        try:
            fcntl.flock(f, fcntl.LOCK_EX)
            json.dump(info.to_dict(), f)
        except IOError as e:
            if e.errno != errno.EAGAIN:
                raise
        finally:
            fcntl.flock(f, fcntl.LOCK_UN)



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false

import os
import platform
import re
import subprocess

from hashlib import md5
from shutil import which
from typing import NamedTuple, Optional, Sequence, Tuple, Union
from urllib.parse import urlparse

from metaflow.exception import MetaflowException
from metaflow.metaflow_config import (
    CONDA_MAGIC_FILE_V2,  # type: ignore
    CONDA_PREFERRED_FORMAT,  # type: ignore
    CONDA_S3ROOT,  # type: ignore
)

from metaflow.metaflow_environment import InvalidEnvironmentException

# NOTA: Most of the code does not assume that there are only two formats BUT the
# transmute code does (since you can only specify the infile -- the outformat and file
# are inferred)
_ALL_CONDA_FORMATS = (".tar.bz2", ".conda")

# List of formats that guarantees the preferred format is first. This is important as
# functions that rely on selecting the "preferred" source of a package rely on the
# preferred format being first.
CONDA_FORMATS = (
    CONDA_PREFERRED_FORMAT,
    *[x for x in _ALL_CONDA_FORMATS if x != CONDA_PREFERRED_FORMAT],
)  # type: Tuple[str, ...]
TRANSMUT_PATHCOMPONENT = "_transmut"


# On Linux systems, called md5sum and md5 on mac
_md5sum_path = which("md5sum") or which("md5")


class CondaException(MetaflowException):
    headline = "Conda ran into an error while setting up environment."

    def __init__(self, error: Union[Sequence[Exception], str]):
        if isinstance(error, list):
            error = "\n".join([str(x) for x in error])
        super(CondaException, self).__init__(error)  # type: ignore


class CondaStepException(CondaException):
    def __init__(self, exception: CondaException, steps: Sequence[str]):
        msg = "Step(s): {steps}, Error: {error}".format(
            steps=steps, error=exception.message
        )
        super(CondaStepException, self).__init__(msg)


def convert_filepath(path: str, file_format: Optional[str] = None) -> Tuple[str, str]:
    if file_format and not path.endswith(file_format):
        for f in CONDA_FORMATS:
            if path.endswith(f):
                path = path[: -len(f)] + file_format
                break
        else:
            raise ValueError(
                "URL '%s' does not end with a supported file format %s"
                % (path, str(CONDA_FORMATS))
            )
    return os.path.split(path)


def get_conda_manifest_path(ds_root: str) -> str:
    return os.path.join(ds_root, CONDA_MAGIC_FILE_V2)  # type: ignore


def get_conda_root(datastore_type: str) -> str:
    if datastore_type == "s3":
        if CONDA_S3ROOT is None:
            # We error on METAFLOW_DATASTORE_SYSROOT_S3 because that is the default used
            raise MetaflowException(msg="METAFLOW_DATASTORE_SYSROOT_S3 must be set!")
        return CONDA_S3ROOT  # type: ignore
    raise CondaException("Unknown datastore type: %s" % datastore_type)


def arch_id() -> str:
    bit = "32"
    if platform.machine().endswith("64"):
        bit = "64"
    if platform.system() == "Linux":
        return "linux-%s" % bit
    elif platform.system() == "Darwin":
        # Support M1 Macmetaf
        if platform.machine() == "arm64":
            return "osx-arm64"
        else:
            return "osx-%s" % bit
    else:
        raise InvalidEnvironmentException(
            "The *@conda* decorator is not supported "
            "outside of Linux and Darwin platforms"
        )


ParseExplicitResult = NamedTuple(
    "ParseExplicitResult",
    [("filename", str), ("url", str), ("url_format", str), ("hash", str)],
)


def parse_explicit_url_conda(url: str) -> ParseExplicitResult:
    # Takes a URL in the form url#hash and returns:
    #  - the filename
    #  - the URL (without the hash)
    #  - the format for the URL
    #  - the hash
    filename = None
    url_format = None

    url_clean, url_hash = url.rsplit("#", 1)

    filename = os.path.split(urlparse(url_clean).path)[1]
    for f in _ALL_CONDA_FORMATS:
        if filename.endswith(f):
            url_format = f
            filename = filename[: -len(f)]
            break
    else:
        raise CondaException(
            "URL '%s' is not a supported format (%s)" % (url, CONDA_FORMATS)
        )
    return ParseExplicitResult(
        filename=filename, url=url_clean, url_format=url_format, hash=url_hash
    )


def parse_explicit_url_pip(url: str) -> ParseExplicitResult:
    # Takes a URL in the form url#hash and returns:
    #  - the filename
    #  - the URL (without the hash)
    #  - the format for the URL
    #  - the hash
    filename = None
    url_format = None

    url_clean, url_hash = url.rsplit("#", 1)
    if not url_hash.startswith("sha256="):
        raise CondaException("URL '%s' has a SHA type which is not supported" % url)
    url_hash = url_hash[7:]

    filename = os.path.split(urlparse(url_clean).path)[1]
    for f in [".whl", ".tar.gz"]:
        if filename.endswith(f):
            url_format = f
            filename = filename[: -len(f)]
            break
    else:
        raise CondaException(
            "URL '%s' is not a supported format (%s)" % (url, str([".whl", ".tar.gz"]))
        )
    return ParseExplicitResult(
        filename=filename, url=url_clean, url_format=url_format, hash=url_hash
    )


def plural_marker(count: int) -> str:
    return "s" if count != 1 else ""



import os
from metaflow.metaflow_config import DATASTORE_SYSROOT_S3
from metaflow.metaflow_config_funcs import from_conf, get_validate_choice_fn


CONDA_S3ROOT = from_conf(
    "CONDA_S3ROOT",
    os.path.join(DATASTORE_SYSROOT_S3, "conda_env") if DATASTORE_SYSROOT_S3 else None,
)

CONDA_MAGIC_FILE_V2 = "conda_v2.cnd"

# Use an alternate dependency resolver for conda packages instead of conda
# Mamba promises faster package dependency resolution times, which
# should result in an appreciable speedup in flow environment initialization.
CONDA_DEPENDENCY_RESOLVER = from_conf(
    "CONDA_DEPENDENCY_RESOLVER", "mamba", get_validate_choice_fn(["mamba", "conda"])
)

# Timeout trying to acquire the lock to create environments
CONDA_LOCK_TIMEOUT = from_conf("CONDA_LOCK_TIMEOUT", 3600)

ENV_PACKAGES_DIRNAME = from_conf("ENV_PACKAGES_DIRNAME", "packages")

# CONDA_REMOTE_INSTALLER_DIRNAME = from_conf(
#     "CONDA_REMOTE_INSTALLER_DIRNAME", "conda-remote"
# )
CONDA_REMOTE_INSTALLER_DIRNAME = from_conf("CONDA_REMOTE_INSTALLER_DIRNAME")

# Binary within CONDA_REMOTE_INSTALLER_DIRNAME to use as the binary on remote instances.
# Use {arch} to specify the architecture. This should be fully functional binary
# CONDA_REMOTE_INSTALLER = from_conf("CONDA_REMOTE_INSTALLER", "conda-{arch}")
CONDA_REMOTE_INSTALLER = from_conf("CONDA_REMOTE_INSTALLER")


# Directory in CONDA_*ROOT that contains the local distribution tarballs. If not specified,
# a conda executable needs to be installed and will be used (if locatable in PATH)
# CONDA_LOCAL_DIST_DIRNAME = from_conf("CONDA_LOCAL_DIST_DIRNAME", "conda-local")
CONDA_LOCAL_DIST_DIRNAME = from_conf("CONDA_LOCAL_DIST_DIRNAME")

# Tar-ball containing the local distribution of conda to use.
# CONDA_LOCAL_DIST = from_conf("CONDA_LOCAL_DIST", "conda-{arch}.tgz")
CONDA_LOCAL_DIST = from_conf("CONDA_LOCAL_DIST")

# Path to the local conda distribution. If this is specified and a conda distribution
# does not exist at this path, Metaflow will attempt to install it using
# CONDA_LOCAL_DIST_DIRNAME and CONDA_LOCAL_DIST
# CONDA_LOCAL_PATH = from_conf("CONDA_LOCAL_PATH", "/usr/local/libexec/metaflow-conda")
CONDA_LOCAL_PATH = from_conf("CONDA_LOCAL_PATH")

# Preferred Format for Conda packages
CONDA_PREFERRED_FORMAT = from_conf(
    "CONDA_PREFERRED_FORMAT",
    ".conda",
    get_validate_choice_fn([".tar.bz2", ".conda"]),
)

# Conda or conda-lock to resolve env
CONDA_PREFERRED_RESOLVER = from_conf(
    "CONDA_PREFERRED_RESOLVER",
    "conda-lock",
    get_validate_choice_fn(["conda", "conda-lock"]),
)

CONDA_DEFAULT_PIP_SOURCES = from_conf("CONDA_DEFAULT_PIP_SOURCES", [])

CONDA_REMOTE_COMMANDS = ("batch", "kubernetes")

DEBUG_OPTIONS = ["conda"]



# pyright: strict, reportTypeCommentUsage=false, reportMissingTypeStubs=false
import os

from functools import partial
from itertools import chain
from typing import Any, Dict, Iterable, Tuple, List, Union, Optional, cast

from metaflow._vendor import click

from metaflow.debug import debug
from metaflow.exception import CommandException
from metaflow.flowspec import FlowSpec

from .conda.conda import Conda
from .conda.conda_step_decorator import get_conda_decorator
from .conda.env_descr import EnvID, PackageSpecification, ResolvedEnvironment
from .conda.terminal_menu import TerminalMenu

# Very simple functions to create some highlights without importing other packages
_BOLD = 1
_ITALIC = 3
_UNDERLINE = 4
_RED = 31
_GREEN = 32
_YELLOW = 33
_BLUE = 34


def _color_str(
    s: str,
    style: Optional[Union[int, List[int]]] = None,
    fg: Optional[int] = None,
    bg: Optional[int] = None,
) -> str:
    if style is None:
        style = []
    elif isinstance(style, int):
        style = [style]
    if fg is not None:
        style.append(fg)
    if bg is not None:
        style.append(bg + 10)
    return "\033[%sm%s\033[0m" % (";".join((str(x) for x in style)), s)


def _get_envs(
    conda: Conda, flow: FlowSpec, skip_known: bool, local_only: bool
) -> List[Tuple[str, Dict[str, ResolvedEnvironment], str, Tuple[str]]]:
    reqs_to_steps = {}  # type: Dict[Tuple[str, Tuple[str]], List[str]]
    for step in flow:
        step_conda_dec = get_conda_decorator(flow, step.__name__)
        # We want to gather steps by the type of requirements they have. This is
        # basically their req ID as well as the requested architectures.
        env_id = step_conda_dec.env_id
        archs = tuple(sorted(step_conda_dec.requested_architectures))
        req_id = env_id.req_id
        reqs_to_steps.setdefault((req_id, archs), []).append(step.__name__)
    # We now look for environments that have the proper req_id and a superset of the
    # architectures asked for
    if debug.conda:
        debug.conda_exec("Requirements to resolve are:")
        for (req_id, archs), steps in reqs_to_steps.items():
            debug.conda_exec(
                "%s (archs: %s) for %s" % (req_id, ", ".join(archs), ", ".join(steps))
            )

    steps_to_potentials = (
        []
    )  # type: List[Tuple[str, Dict[str, ResolvedEnvironment], str, Tuple[str]]]
    for (req_id, archs), steps in reqs_to_steps.items():
        # We can pick just a single arch to look for potentials; we know we need
        # at least that one
        arch = archs[0]
        need_archs = set(archs)
        possible_envs = conda.environments(req_id, arch, local_only=local_only)
        possible_envs = [
            r for r in possible_envs if need_archs.issubset(r[1].co_resolved_archs)
        ]
        if skip_known:
            default_env_id = conda.get_default_environment(req_id, arch)
            if default_env_id and default_env_id in [env[0] for env in possible_envs]:
                debug.conda_exec(
                    "Have known local default for %s -- skipping" % ", ".join(steps)
                )
                continue
        steps_to_potentials.append(
            (
                ", ".join(steps),
                {env_id.full_id: env for env_id, env in possible_envs},
                req_id,
                archs,
            )
        )
    return steps_to_potentials


@click.group()
def cli():
    pass


@cli.group(help="Commands related to environment manipulation.")
@click.pass_obj
def environment(obj):
    if obj.environment.TYPE != "conda":
        raise CommandException("'environment' requires a Conda environment")
    pass


@environment.command(help="Show available environments for steps in a flow")
@click.option(
    "--local-only/--no-local-only",
    show_default=True,
    default=False,
    help="Search only locally cached environments",
)
@click.option(
    "--menu/--no-menu",
    show_default=True,
    default=True,
    help="Use a menu to display the environment or print",
)
@click.pass_obj
def list(obj: Any, local_only: bool = True, menu: bool = False):
    conda = Conda(obj.echo, obj.flow_datastore.TYPE)
    steps_to_potentials = _get_envs(
        conda, obj.flow, skip_known=False, local_only=local_only
    )
    for steps_query, potentials, req_id, _ in steps_to_potentials:
        local_instances = _local_instances_for_req_id(conda, req_id)
        plural = "s" if len(steps_query.split(",", 1)) > 1 else ""
        if len(potentials) == 0:
            if not obj.is_quiet:
                obj.echo(
                    "No %sknown environments for step%s %s"
                    % (("locally " if local_only else ""), plural, steps_query)
                )
            continue
        if menu:
            m = TerminalMenu(
                menu_entries=(
                    _format_resolved_env_for_menu(conda, env)
                    for env in potentials.values()
                ),
                title="Available environments for step%s %s (hash: %s)"
                % (plural, steps_query, req_id),
                #% _color_str(steps_query, style=_BOLD, fg=_RED),
                preview_title="Packages and local instances",
                column_headers=_columns_for_menu(),
                preview_cache_result=True,
                preview_border=True,
                preview_command=partial(_info_for_env, potentials, local_instances),
                preview_size=0.5,
            )
            m.show()
        elif obj.is_quiet:
            print(steps_query)
            for env in potentials.values():
                _quiet_print_env(obj, env, local_instances.get(env.env_id.full_id))
        else:
            obj.echo(
                "### Environments for step%s *%s* (hash: %s) ###"
                % (plural, steps_query, req_id)
            )
            first_env = True
            for env in potentials.values():
                if not first_env:
                    obj.echo("\n*~~~*\n")
                first_env = False
                _pretty_print_env(
                    obj,
                    env,
                    env.env_id
                    == conda.get_default_environment(
                        env.env_id.req_id, env.env_id.arch
                    ),
                    local_instances.get(env.env_id.full_id),
                )
            obj.echo("\n\n")


@environment.command(help="Select resolved environments for steps in a flow")
@click.option(
    "--skip-known/--no-skip-known",
    show_default=True,
    default=True,
    help="Skip steps with a locally known resolved environment",
)
@click.option(
    "--local-only/--no-local-only",
    show_default=True,
    default=False,
    help="Search only locally cached environments",
)
@click.option(
    "--dry-run/--no-dry-run",
    show_default=True,
    default=False,
    help="Dry-run -- do not update cached environments",
)
@click.pass_obj
def select_resolved(
    obj: Any, skip_known: bool = True, local_only: bool = False, dry_run: bool = False
):

    conda = Conda(obj.echo, obj.flow_datastore.TYPE)
    steps_to_potentials = _get_envs(
        conda, obj.flow, skip_known=skip_known, local_only=local_only
    )
    for steps_query, potentials, req_id, archs in steps_to_potentials:
        local_instances = _local_instances_for_req_id(conda, req_id)
        plural = "s" if len(steps_query.split(",", 1)) > 1 else ""
        menu = TerminalMenu(
            menu_entries=chain(
                [_unset_env_for_menu()],
                (
                    _format_resolved_env_for_menu(conda, env)
                    for env in potentials.values()
                ),
            ),
            title="Select environment for step%s %s" % (plural, steps_query),
            #% _color_str(steps_query, style=_BOLD, fg=_RED),
            preview_title="Packages",
            column_headers=_columns_for_menu(),
            preview_cache_result=True,
            preview_border=True,
            preview_command=partial(_info_for_env, potentials, local_instances),
            preview_size=0.5,
        )
        idx = None
        while idx is None:
            idx = menu.show()
        idx = cast(int, idx)
        if idx == 0:
            # This is the unset option, so we clear the default
            for arch in archs:
                conda.clear_default_environment(req_id, arch)
            if dry_run and not obj.is_quiet:
                obj.echo(
                    "Clearing default environment for step%s %s" % (plural, steps_query)
                )
        else:
            # This is a set option so we set the default to that value for all
            # requested archs
            full_id = tuple(potentials.keys())[idx - 1]

            for arch in archs:
                conda.set_default_environment(
                    EnvID(req_id=req_id, full_id=full_id, arch=arch)
                )
            if dry_run and not obj.is_quiet:
                obj.echo(
                    "Setting default environment to %s for step%s %s"
                    % (full_id, plural, steps_query)
                )

    if not dry_run:
        conda.write_out_environments()


def _info_for_env(
    env_dict: Dict[str, ResolvedEnvironment],
    local_instances: Dict[str, List[str]],
    full_id: str,
) -> List[str]:
    resolved_env = env_dict.get(full_id)
    if resolved_env is None:
        return [""]
    to_return = []  # type: List[str]
    pip_packages = []  # type: List[PackageSpecification]
    conda_packages = []  # type: List[PackageSpecification]
    for p in resolved_env.packages:
        if p.TYPE == "pip":
            pip_packages.append(p)
        else:
            conda_packages.append(p)

    pip_packages.sort(key=lambda x: x.package_name)
    conda_packages.sort(key=lambda x: x.package_name)
    local_dirs = local_instances.get(full_id)
    if local_dirs:
        to_return.append(
            _color_str("Locally Present", style=[_UNDERLINE, _BOLD], bg=_YELLOW)
        )
        to_return.extend(
            (
                "- %s"
                % d.replace(resolved_env.env_id.req_id, "<req-id>").replace(
                    full_id, "<id>"
                )
                for d in local_dirs
            )
        )
    if conda_packages:
        to_return.append(
            _color_str("Conda Packages", style=[_UNDERLINE, _BOLD], bg=_YELLOW)
        )
        to_return.extend(
            (
                _color_str("- %s==%s" % (p.package_name, p.package_version), fg=_GREEN)
                for p in conda_packages
            )
        )
    if pip_packages:
        to_return.append(
            _color_str("Pip Packages", style=[_UNDERLINE, _BOLD], bg=_YELLOW)
        )
        to_return.extend(
            (
                _color_str("- %s==%s" % (p.package_name, p.package_version), fg=_BLUE)
                for p in pip_packages
            )
        )
    return to_return


def _columns_for_menu() -> List[str]:
    return ["Env ID", "Archs", "Creator", "Resolved On", "Tags"]


def _unset_env_for_menu() -> List[str]:
    return ["re-resolve", "", "", "", ""]


def _format_resolved_env_for_menu(conda: Conda, env: ResolvedEnvironment) -> List[str]:
    # Returns full_id, resolved_archs, resolved_by, resolved_on, tags
    is_default = env.env_id == conda.get_default_environment(
        env.env_id.req_id, env.env_id.arch
    )
    return [
        "%s|%s" % (env.env_id.full_id, env.env_id.full_id),
        ", ".join(env.co_resolved_archs),
        env.resolved_by,
        env.resolved_on.isoformat(),
        ("[default]" if is_default else ""),
    ]


def _local_instances_for_req_id(conda: Conda, req_id: str) -> Dict[str, List[str]]:
    created_envs = conda.created_environments()  # This returns all MF environments
    local_instances = {}  # type: Dict[str, List[str]]
    for env_id, present_list in created_envs.items():
        if env_id.req_id != req_id:
            continue
        local_instances[env_id.full_id] = [os.path.basename(p) for p in present_list]
    return local_instances


def _quiet_print_env(
    obj, env: ResolvedEnvironment, local_instances: Optional[List[str]]
):
    env_id = env.env_id

    pip_packages = []  # type: List[PackageSpecification]
    conda_packages = []  # type: List[PackageSpecification]
    for p in env.packages:
        if p.TYPE == "pip":
            pip_packages.append(p)
        else:
            conda_packages.append(p)

    pip_packages.sort(key=lambda x: x.package_name)
    conda_packages.sort(key=lambda x: x.package_name)
    obj.echo_always(
        "%s %s %s %s %s %s %s %s conda:%s pip:%s %s"
        % (
            env_id.req_id,
            env_id.full_id,
            env_id.arch,
            env.resolved_on.isoformat(),
            env.resolved_by,
            ",".join(env.co_resolved_archs),
            ",".join([str(d) for d in env.deps]),
            ",".join([str(s) for s in env.sources]) if env.sources else "NONE",
            ",".join(
                ["%s==%s" % (p.package_name, p.package_version) for p in conda_packages]
            ),
            ",".join(
                ["%s==%s" % (p.package_name, p.package_version) for p in pip_packages]
            ),
            ",".join(local_instances) if local_instances else "NONE",
        )
    )


def _pretty_print_env(
    obj,
    env: ResolvedEnvironment,
    is_default: bool,
    local_instances: Optional[List[str]],
):

    pip_packages = []  # type: List[PackageSpecification]
    conda_packages = []  # type: List[PackageSpecification]
    for p in env.packages:
        if p.TYPE == "pip":
            pip_packages.append(p)
        else:
            conda_packages.append(p)

    pip_packages.sort(key=lambda x: x.package_name)
    conda_packages.sort(key=lambda x: x.package_name)
    obj.echo(
        "*%sEnvironment full hash* %s\n"
        % ("DEFAULT " if is_default else "", env.env_id.full_id)
    )
    obj.echo("*Arch* %s" % env.env_id.arch)
    obj.echo("*Available on* %s\n" % ", ".join(env.co_resolved_archs))
    obj.echo("*Resolved on* %s" % env.resolved_on)
    obj.echo("*Resolved by* %s\n" % env.resolved_by)
    if local_instances:
        obj.echo("*Locally present as* %s\n" % ", ".join(local_instances))

    obj.echo("*User-requested packages* %s" % ", ".join([str(d) for d in env.deps]))
    obj.echo("*User sources* %s" % ", ".join([str(s) for s in env.sources]))
    obj.echo(
        "*Conda Packages installed* %s"
        % ", ".join(
            ["%s==%s" % (p.package_name, p.package_version) for p in conda_packages]
        )
    )
    obj.echo(
        "*Pip Packages installed* %s"
        % ", ".join(
            ["%s==%s" % (p.package_name, p.package_version) for p in pip_packages]
        )
    )
